# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimString,
    Segment,
)


class EpicenterSegmentH(HeaderSegment):
    sender_name = HprimString(5)
    version = HprimString(13, default="V1.0")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class EpicenterSegmentP(Segment):
    patient_id = HprimString(4)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    sex = HprimString(9)
    diagnostic = HprimString(19)
    admission_date = HprimDate(24)
    room = HprimString(26)
    location = HprimString(33)

    orders = HprimForeign("O", "patient")


class EpicenterSegmentQ(Segment):
    starting_accession_number = HprimString((3, 1))
    accession_number = HprimString((3, 2))
    sequence_number = HprimString((3, 3))
    test_id = HprimString((5, 1))
    test_status = HprimString((5, 2))
    result_qualifier = HprimString((5, 5))
    time_qualifier = HprimDate(6)
    starting_date = HprimDate(7)
    end_date = HprimDate(8)


class EpicenterSegmentO(Segment):
    accession_number = HprimString((3, 1))
    isolate_number = HprimString((3, 2))
    organism = HprimString((3, 3))
    test_id = HprimString((5, 4))
    test_consumable_number = HprimString((5, 5))
    priority = HprimString(6)
    collection_date = HprimDate(8)
    receipt_date = HprimDate(15)
    specimen_type = HprimString(16)
    physician_id = HprimString(17)
    report_type = HprimString(26)

    results = HprimForeign("R", "order")


class EpicenterSegmentR(Segment):
    result_type_code = HprimString((3, 4))
    test_number = HprimString((3, 5))
    antibiotic = HprimString((3, 6))
    test_status = HprimString((4, 1))
    result_date_mic = HprimString((4, 2))
    organism_id = HprimString((4, 3))
    resistance_marker1 = HprimString((4, 4))
    resistance_marker2 = HprimString((4, 5))
    resistance_marker3 = HprimString((4, 6))
    resistance_marker4 = HprimString((4, 7))
    resistance_marker5 = HprimString((4, 8))
    resistance_marker6 = HprimString((4, 9))
    resistance_marker7 = HprimString((4, 10))
    resistance_marker8 = HprimString((4, 11))
    resistance_marker9 = HprimString((4, 12))
    resistance_marker10 = HprimString((4, 13))
    status = HprimString(9)
    test_start_date = HprimDate(12)
    result_status_date = HprimList(13)
    instrument_type = HprimString((14, 1))
    instrument_number = HprimString((14, 2))
    instrument_location = HprimString((14, 5))


class EpicenterSegmentC(Segment):
    comment_text = HprimString(4)
    # S: commentaire specimen
    # I: commentaire isolat
    comment_type = HprimString(5)


class EpicenterSegmentL(Segment):
    code = HprimString(3)


class Driver(AstmMessageV2):
    SPEC = {"code": "epicenter", "label": "Epicenter PHOENIX BD"}

    SEGMENT_CLASS_MAP = {
        "H": EpicenterSegmentH,
        "Q": EpicenterSegmentQ,
        "P": EpicenterSegmentP,
        "O": EpicenterSegmentO,
        "R": EpicenterSegmentR,
        "C": EpicenterSegmentC,
        "L": EpicenterSegmentL,
    }

    DEFAULT_SEPARATORS = "|\\^&"

    OPTIONS = [
        AstmMessageV2.Option(
            "epicenter_liste_ignorer_galerie",
            label="Liste des codes de galerie à ne pas envoyer avec la demande",
        ),
        AstmMessageV2.Option(
            "epicenter_hour_min_sep",
            label="Le séparateur entre les heures et les minutes",
            default="H",
        ),
        AstmMessageV2.Option(
            "epicenter_liste_resistance_marker",
            label="Transcodage des resistances marker des champs R.4.4 jusqu'à R.4.13 selon le format <resistance_marker1>=<equivalence1>/<resistance_marker2>=<equivalence2>",
            default="",
        ),
        AstmMessageV2.BooleanOption(
            "epicenter_liste_resistance_marker_activation",
            label="Activation du transcodage des phénotypes par le driver",
            default=False,
        ),
        AstmMessageV2.IntegerOption(
            "epicenter_test_start_date",
            label="Nombre de minutes pour retarder la date du résultat (champ R.12)",
            default=0,
        ),
    ]

    def __init__(
        self,
        epicenter_liste_ignorer_galerie="",
        epicenter_hour_min_sep="H",
        epicenter_liste_resistance_marker="",
        epicenter_liste_resistance_marker_activation=False,
        epicenter_collection_date_minute=0,
        epicenter_test_start_date=0,
        **kwargs
    ):
        self.epicenter_liste_ignorer_galerie = epicenter_liste_ignorer_galerie
        self.epicenter_hour_min_sep = epicenter_hour_min_sep
        self.epicenter_liste_resistance_marker = epicenter_liste_resistance_marker
        self.epicenter_liste_resistance_marker_activation = (
            epicenter_liste_resistance_marker_activation
        )
        self.epicenter_test_start_date = epicenter_test_start_date
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = EpicenterSegmentH(type_segment="H", **kwargs)
        message.end = EpicenterSegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        msg = self.create(date=context["now"])
        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: None},
        ):
            identification_germ_to_send = False
            for sample, action in patient.samples:
                for result in sample["results"]:
                    if result["resu_json"] and "germes" in result["resu_json"]:
                        identification_germ_to_send = True
                if identification_germ_to_send:
                    # Identification de germe à envoyer
                    # Une identification par segment patient
                    for result in sample["results"]:
                        if result["resu_json"] and "germes" in result["resu_json"]:
                            for resu_json_germe in result["resu_json"]["germes"]:
                                if not isinstance(resu_json_germe, dict):
                                    continue
                                try:
                                    rank_germ = resu_json_germe.get("rang")
                                except KeyError:
                                    rank_germ = ""
                                try:
                                    key_germ = [
                                        cle
                                        for cle in resu_json_germe.keys()
                                        if cle.lower() == "id"
                                    ]
                                    id_germ = resu_json_germe.get(
                                        key_germ[0] if key_germ else None
                                    )
                                    if isinstance(id_germ, list):
                                        id_germ = ""
                                except KeyError:
                                    id_germ = ""
                                try:
                                    key_nom_atb = [
                                        cle
                                        for cle in resu_json_germe.keys()
                                        if cle.lower() == "nom_antibiogramme"
                                    ]
                                    nom_atb = resu_json_germe.get(
                                        key_nom_atb[0] if key_nom_atb else None
                                    )
                                except KeyError:
                                    nom_atb = ""
                                ignore_panel = False
                                if nom_atb:
                                    ignore_panel = (
                                        nom_atb in self.epicenter_liste_ignorer_galerie
                                    )
                                seg_p = msg.patients.add(
                                    patient_id=patient.patient_id,
                                    last_name=patient.last_name,
                                    first_name=patient.first_name,
                                    birthdate=patient.birthdate,
                                    sex=patient.sex,
                                    location=patient.location,
                                )
                                seg_p.room = sample.room
                                seg_p.admission_date = sample.admission_date
                                seg_p.diagnostic = sample.clinical_information
                                # Envoi du germe identifié avec le rang
                                for order in sample.orders:
                                    seg_o = seg_p.orders.add(
                                        accession_number=sample.sample_id,
                                        isolate_number="1"
                                        if rank_germ is None
                                        else str(rank_germ),
                                        organism=id_germ[2:]
                                        if id_germ and id_germ.startswith("G_")
                                        else id_germ,
                                        test_id=order.code,
                                        collection_date=sample.collection_date,
                                        specimen_type=sample.specimen,
                                    )
                                    # Une demande par nature de prélèvement et
                                    # par échantillon
                                    break
                                physician = [
                                    p
                                    for p in patient.physicians
                                    if p.type in ("prescriber", "attending")
                                ]
                                seg_o.physician_id = (
                                    physician[0].code if physician else ""
                                )
                                if self.epicenter_test_start_date:
                                    test_start_date = result.date - datetime.timedelta(
                                        minutes=self.epicenter_test_start_date
                                    )
                                else:
                                    test_start_date = result.date
                                seg_o.results.add(
                                    result_type_code="ID",
                                    test_status="MALDI",
                                    result_date_mic=id_germ[2:]
                                    if id_germ and id_germ.startswith("G_")
                                    else id_germ,
                                    status="F",
                                    test_start_date=test_start_date,
                                )
                                # Envoi du nom de l'antibiogramme (code panel) avec le rang
                                # du germe
                                if nom_atb and not ignore_panel:
                                    for order in sample.orders:
                                        seg_o = seg_p.orders.add(
                                            accession_number=sample.sample_id,
                                            isolate_number="1"
                                            if rank_germ is None
                                            else str(rank_germ),
                                            test_id=nom_atb,
                                        )
                                    seg_o.results.add(
                                        result_type_code="AST-CMI",
                                        test_status="Pending",
                                        status="P",
                                    )
                else:
                    # Demande de prélèvement classique sans identification à
                    # envoyer
                    if len(sample.orders) == 1:
                        if sample.orders[0].code == "MBT_ID":
                            return self.create(date=context["now"])
                    seg_p = msg.patients.add(
                        patient_id=patient.patient_id,
                        last_name=patient.last_name,
                        first_name=patient.first_name,
                        birthdate=patient.birthdate,
                        sex=patient.sex,
                        location=patient.location,
                    )
                    seg_p.room = sample.room
                    seg_p.admission_date = sample.admission_date
                    seg_p.diagnostic = sample.clinical_information
                    for order in sample.orders:
                        if order.code == "MBT_ID":
                            continue
                        if order.code == sample.specimen or order.code.startswith(
                            "AST-CMI"
                        ):
                            # Pas d'envoi de code demande pour les demandes
                            # d'ATB
                            order_code = ""
                        else:
                            # Cela concerne les autres demandes comme celles
                            # d'hémoculture
                            order_code = order.code
                        seg_o = seg_p.orders.add(
                            accession_number=sample.sample_id,
                            test_id=order_code,
                            collection_date=sample.collection_date,
                            specimen_type=sample.specimen,
                        )
                    physician = [
                        p
                        for p in patient.physicians
                        if p.type in ("prescriber", "attending")
                    ]
                    seg_o.physician_id = physician[0].code if physician else ""
        if msg.patients:
            return msg

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                worklist.get_or_create(type="patient", sample_id=seg_q.accession_number)
        else:
            worklist = Worklist(context="result")
            for seg_p in message.patients:
                for seg_o in seg_p.orders:
                    sample = worklist.get_or_create(
                        type="patient", sample_id=seg_o.accession_number
                    )
                    commentaires = []
                    commentaire_seg_c = []
                    for seg_c in seg_o.commentaires:
                        # Ignorer les commentaires de type E
                        if seg_c.comment_type == "E":
                            continue
                        # Si l'option n'est pas activée, traiter tous les
                        # commentaires codés ou non
                        commentaire_seg_c.append(seg_c.comment_text)
                    # Résulat d'hémoculture
                    liste_atb = []
                    blood_culture_result = False
                    resistance_marker = []
                    for seg_r in seg_o.results:
                        if seg_r.result_status_date:
                            date = datetime.datetime.strptime(
                                seg_r.result_status_date[0], "%Y%m%d%H%M%S"
                            )
                        else:
                            date = None
                        if seg_r.result_type_code.startswith("GND"):
                            instrument = seg_r.instrument_type
                            if seg_r.instrument_type and seg_r.instrument_number:
                                instrument = "{}-{}".format(
                                    seg_r.instrument_type, seg_r.instrument_number
                                )
                            blood_culture_result = True
                            sample.results.add(
                                code="{test_id}{result_type}".format(
                                    test_id=seg_o.test_id,
                                    result_type=seg_r.result_type_code,
                                ),
                                value=seg_r.test_status,
                                comments=commentaires,
                                date=date,
                                instrument=instrument,
                            )
                            if seg_r.test_number:
                                sample.results.add(
                                    code="BOTTLE_ID",
                                    value=seg_r.test_number,
                                    date=date,
                                    instrument=instrument,
                                )
                            if seg_r.instrument_location:
                                sample.results.add(
                                    code="BOTTLE_RACK",
                                    value=seg_r.instrument_location,
                                    date=date,
                                    instrument=instrument,
                                )
                            if instrument:
                                sample.results.add(
                                    code="INSTRUMENT", value=instrument, date=date
                                )
                            # Date et heure d'introduction du flacon
                            if seg_r.test_start_date:
                                sample.results.add(
                                    code="{test_id}INTRODUCE_DATETIME".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=seg_r.test_start_date.strftime(
                                        "%Y/%m/%d %H:%M:%S"
                                    ),
                                )
                                sample.results.add(
                                    code="{test_id}BOTTLE_INTRODUCE_DATETIME".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=seg_r.test_start_date.strftime(
                                        "%Y%m%d%H%M%S"
                                    ),
                                )
                                sample.results.add(
                                    code="{test_id}INTRODUCE_DATE".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=seg_r.test_start_date.strftime("%Y/%m/%d"),
                                )
                                sample.results.add(
                                    code="{test_id}INTRODUCE_TIME".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=seg_r.test_start_date.strftime("%H:%M:%S"),
                                )
                                sample.results.add(
                                    code="{test_id}INTRODUCTION_DATE".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=seg_r.test_start_date.strftime("%d/%m/%Y"),
                                )
                                sample.results.add(
                                    code="{test_id}INTRODUCTION_HEURE".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=seg_r.test_start_date.strftime("%H:%M"),
                                )

                            if seg_r.result_status_date:
                                # Heure et date du retrait du flacon
                                date_result_determinated = datetime.datetime.strptime(
                                    seg_r.result_status_date[0], "%Y%m%d%H%M%S"
                                )
                                sample.results.add(
                                    code="{test_id}DETERMINED_DATETIME".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=date_result_determinated.strftime(
                                        "%Y%m%d%H%M%S"
                                    ),
                                )
                                sample.results.add(
                                    code="{test_id}DETERMINED_DATE".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=date_result_determinated.strftime("%Y/%m/%d"),
                                )
                                sample.results.add(
                                    code="{test_id}DETERMINED_TIME".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=date_result_determinated.strftime("%H:%M:%S"),
                                )
                                sample.results.add(
                                    code="{test_id}DATE_DETERMINATION".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=date_result_determinated.strftime("%d/%m/%Y"),
                                )
                                sample.results.add(
                                    code="{test_id}HEURE_DETERMINATION".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=date_result_determinated.strftime("%H:%M"),
                                )
                                growth_delay = (
                                    date_result_determinated - seg_r.test_start_date
                                )
                                sample.results.add(
                                    code="{test_id}GROWTH_TIME".format(
                                        test_id=seg_o.test_id
                                    ),
                                    value=self.timedelta_to_hour_min(
                                        growth_delay, sep=self.epicenter_hour_min_sep
                                    ),
                                )
                                if "positive" in seg_r.test_status.lower():
                                    # Date et heure de positivité
                                    sample.results.add(
                                        code="{test_id}POSITIVITY_DATETIME".format(
                                            test_id=seg_o.test_id
                                        ),
                                        value=date_result_determinated.strftime(
                                            "%Y%m%d%H%M%S"
                                        ),
                                    )
                                    # Date de positivité
                                    sample.results.add(
                                        code="{test_id}POSITIVITY_DATE".format(
                                            test_id=seg_o.test_id
                                        ),
                                        value=date_result_determinated.strftime(
                                            "%Y/%m/%d"
                                        ),
                                    )
                                    sample.results.add(
                                        code="{test_id}POSITIVITE_DATE".format(
                                            test_id=seg_o.test_id
                                        ),
                                        value=date_result_determinated.strftime(
                                            "%d/%m/%Y"
                                        ),
                                    )
                                    # Heure de positivité
                                    sample.results.add(
                                        code="{test_id}POSITIVITY_TIME".format(
                                            test_id=seg_o.test_id
                                        ),
                                        value=date_result_determinated.strftime(
                                            "%H:%M:%S"
                                        ),
                                    )
                                    sample.results.add(
                                        code="{test_id}POSITIVITE_HEURE".format(
                                            test_id=seg_o.test_id
                                        ),
                                        value=date_result_determinated.strftime(
                                            "%H:%M"
                                        ),
                                    )
                                    # Delai de positivité
                                    if seg_r.result_status_date:
                                        delay_positivity = (
                                            date_result_determinated
                                            - seg_r.test_start_date
                                        )
                                        sample.results.add(
                                            code="TIME_TO_DETECTION_POSITIVITY_{test_id}".format(
                                                test_id=seg_o.test_id
                                            ),
                                            value=self.timedelta_to_hour_min(
                                                delay_positivity,
                                                sep=self.epicenter_hour_min_sep,
                                            ),
                                        )
                        elif seg_r.result_type_code == "AST":
                            liste_atb.append(
                                {
                                    "antibiotic": "A_{antibiotic}".format(
                                        antibiotic=seg_r.antibiotic
                                    ),
                                    "mic": seg_r.result_date_mic,
                                    "sensitivity": seg_r.organism_id,
                                }
                            )
                        elif seg_r.result_type_code == "STREAK":
                            sample.results.add(
                                code="{result_type}_{specimen_type}".format(
                                    result_type=seg_r.result_type_code,
                                    specimen_type=seg_o.specimen_type,
                                ),
                                value=seg_r.test_status,
                            )
                        elif seg_r.result_type_code == "ID":
                            resistance_marker = [
                                marker
                                for marker in (
                                    seg_r.resistance_marker1,
                                    seg_r.resistance_marker2,
                                    seg_r.resistance_marker3,
                                    seg_r.resistance_marker4,
                                    seg_r.resistance_marker5,
                                    seg_r.resistance_marker6,
                                    seg_r.resistance_marker7,
                                    seg_r.resistance_marker8,
                                    seg_r.resistance_marker9,
                                    seg_r.resistance_marker10,
                                )
                                if marker
                            ]
                            if self.epicenter_liste_resistance_marker_activation:
                                for (
                                    phenotype
                                ) in self.epicenter_liste_resistance_marker.split("/"):
                                    if "=" in phenotype:
                                        phenotype_code, phenotype_commentaire = phenotype.split(
                                            "="
                                        )
                                        if phenotype_code in resistance_marker:
                                            commentaires.append(phenotype_commentaire)

                    if commentaire_seg_c:
                        commentaires.extend(commentaire_seg_c)
                    if liste_atb:
                        value = [
                            {
                                "sub_code": seg_o.isolate_number,
                                "antibiogram": liste_atb,
                                "value": "G_{}".format(seg_o.organism),
                            }
                        ]
                        if (
                            not self.epicenter_liste_resistance_marker_activation
                            and resistance_marker
                        ):
                            value[0]["phenotypes"] = resistance_marker
                        sample.results.add(
                            code=seg_o.specimen_type,
                            value=value,
                            date=seg_o.receipt_date,
                            nom_antibiogramme=seg_r.resistance_marker3,
                            comments=commentaires,
                        )
                    if (
                        not blood_culture_result
                        and seg_o.specimen_type
                        and seg_o.isolate_number
                    ):
                        # version classique avec réponse automate
                        sample.results.add(
                            code="GERME_{isolate_number}_{specimen_type}".format(
                                isolate_number=seg_o.isolate_number,
                                specimen_type=seg_o.specimen_type,
                            ),
                            value="G_{organism}".format(organism=seg_o.organism),
                        )
                        # version avec complexe pour l'identification du germe
                        sample.results.add(
                            code="IDENT_{specimen_type}".format(
                                specimen_type=seg_o.specimen_type
                            ),
                            value={
                                "seq": "0",
                                "type_resultat": "identification_germes",
                                "germes": [
                                    {
                                        "rang": int(seg_o.isolate_number),
                                        "id": "G_{organism}".format(
                                            organism=seg_o.organism
                                        ),
                                    }
                                ],
                            },
                        )
        return worklist

    @staticmethod
    def timedelta_to_hour_min(timedelta, sep=""):
        if isinstance(timedelta, datetime.timedelta):
            seconds = timedelta.total_seconds()
        else:
            seconds = int(timedelta)
        hours = int(seconds // 3600)
        minutes = int(seconds // 60 % 60)
        return "{0:02d}{1}{2:02d}".format(hours, sep, minutes)
